@@ -52,11 +52,11 @@
 			* zero, the API is not considered stable
 			* and can change without a bump of the
 			* major version). */
-#define MINVERSION 2   /* API compatible, ABI may change,
+#define MINVERSION 3   /* API compatible, ABI may change,
 			* functional enhancements only, consumer
 			* can be left unchanged if enhancements are
 			* not considered. */
-#define PATCHLEVEL 2   /* API / ABI compatible, no functional
+#define PATCHLEVEL 0   /* API / ABI compatible, no functional
 			* changes, no enhancements, bug fixes
 			* only. */
 
@@ -228,37 +228,15 @@ static int drng_chacha20_selftest(void)
 	return drng_chacha20_selftest_one(&chacha20, &expected[0]);
 }
 
-/******************************** Seed Source ********************************/
-
-#ifdef GETRANDOM	/* getrandom system call */
+/********************* getrandom system call seed source *********************/
+#ifdef GETRANDOM
 
 #include <limits.h>
-
-struct seed_source {
-#define OVERSAMPLINGRATE 1
-	void *unused;
-};
-
-static int drng_seedsource_alloc(struct seed_source *source)
-{
-	(void)source;
-	return 0;
-}
-
-static void drng_seedsource_dealloc(struct seed_source *source)
-{
-	(void)source;
-	return;
-}
-
-static int drng_get_seed(struct seed_source *source, uint8_t *buf,
-			 uint32_t buflen)
+static int drng_getrandom_get(uint8_t *buf, uint32_t buflen)
 {
 	uint32_t len = 0;
 	ssize_t ret;
 
-	(void)source;
-
 	if (buflen > INT_MAX)
 		return 0;
 
@@ -272,62 +250,127 @@ static int drng_get_seed(struct seed_source *source, uint8_t *buf,
 	return len;
 }
 
-#elif JENT		/* CPU execution time jitter RNG */
-#include "jitterentropy.h"
-
-struct seed_source {
-#define OVERSAMPLINGRATE 2
-	struct rand_data *ec;
-};
-
-static int drng_seedsource_alloc(struct seed_source *source)
+#else
+static int drng_getrandom_get(uint8_t *buf, uint32_t buflen)
 {
-	int ret = jent_entropy_init();
-
-	if (ret)
-		return ret;
-
-	source->ec = jent_entropy_collector_alloc(0, 0);
-	if (!source->ec)
-		return 1;
+	(void)buf;
+	(void)buflen;
 
 	return 0;
 }
 
-static void drng_seedsource_dealloc(struct seed_source *source)
+#endif
+
+/*************************** Jitter RNG seed source ***************************/
+#ifdef JENT
+
+#include "jitterentropy.h"
+
+struct jent_noise_source {
+	struct rand_data *ec;
+	int initialized;
+};
+
+static struct jent_noise_source jent_noise_source = {
+	NULL,
+	0,
+};
+
+static int drng_jent_alloc()
 {
-	jent_entropy_collector_free(source->ec);
-	source->ec = NULL;
+	int ret = jent_entropy_init();
+
+	if (ret) {
+		jent_noise_source.initialized = -1;
+		return -EFAULT;
+	}
+
+	jent_noise_source.ec = jent_entropy_collector_alloc(0, 0);
+	if (!jent_noise_source.ec)
+		return -ENOMEM;
+
+	return 0;
 }
 
-static int drng_get_seed(struct seed_source *source, uint8_t *buf,
-			 uint32_t buflen)
+static void drng_jent_dealloc(void)
 {
-	return jent_read_entropy(source->ec, (char *)buf, buflen);
+	if (jent_noise_source.initialized != 1)
+		return;
+
+	jent_entropy_collector_free(jent_noise_source.ec);
+	jent_noise_source.ec = NULL;
+	jent_noise_source.initialized = 0;
 }
 
-#elif DEVRANDOM		/* Reading of /dev/random */
+static int drng_jent_get(uint8_t *buf, uint32_t buflen)
+{
+	if (!jent_noise_source.initialized) {
+		int ret = drng_jent_alloc();
+
+		if (ret)
+			return ret;
+
+		jent_noise_source.initialized = 1;
+	}
+
+	if (jent_noise_source.initialized > 0)
+		return jent_read_entropy(jent_noise_source.ec,
+					 (char *)buf, buflen);
+
+	return 0;
+}
+
+#else
+static int drng_jent_get(uint8_t *buf, uint32_t buflen)
+{
+	(void)buf;
+	(void)buflen;
+
+	return 0;
+}
+
+static void drng_jent_dealloc(void)
+{
+	return;
+}
+#endif
+
+/************************** /dev/random seed source **************************/
+#ifdef DEVRANDOM
+
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <limits.h>
 
-struct seed_source {
-#define OVERSAMPLINGRATE 1
-	int fd;
-};
+static int random_fd = -1;
 
-static int drng_get_seed(struct seed_source *source, uint8_t *buf,
-			 uint32_t buflen)
+static int drng_random_alloc(void)
+{
+	random_fd = open("/dev/random", O_RDONLY|O_CLOEXEC);
+	if (0 > random_fd)
+		return -EBADFD;
+
+	return 0;
+}
+
+static int drng_random_get(uint8_t *buf, uint32_t buflen)
 {
 	uint32_t len = 0;
 	ssize_t ret;
 
+	if (random_fd == -1) {
+		int ret = drng_random_alloc();
+
+		if (ret)
+			return ret;
+	}
+
 	if (buflen > INT_MAX)
 		return 0;
 
 	do {
-		ret = read(source->fd, (buf + len), (buflen - len));
+		ret = read(random_fd, (buf + len), (buflen - len));
 		if (0 < ret)
 			len += ret;
 	} while ((0 < ret || EINTR == errno || ERESTART == errno)
@@ -336,30 +379,34 @@ static int drng_get_seed(struct seed_source *source, uint8_t *buf,
 	return len;
 }
 
-static int drng_seedsource_alloc(struct seed_source *source)
+static void drng_random_dealloc(void)
 {
-	source->fd = open("/dev/random", O_RDONLY|O_CLOEXEC);
-	if (0 > source->fd)
-		return 1;
+	if (random_fd < 0)
+		return;
+
+	close(random_fd);
+	random_fd = -1;
+}
+
+#else
+static int drng_random_get(uint8_t *buf, uint32_t buflen)
+{
+	(void)buf;
+	(void)buflen;
 
 	return 0;
 }
 
-static void drng_seedsource_dealloc(struct seed_source *source)
+static void drng_random_dealloc(void)
 {
-	close(source->fd);
-	source->fd = -1;
+	return;
 }
-
-#else
-#error "No seed source defined"
 #endif
 
 /******************************* ChaCha20 DRNG *******************************/
 
 struct chacha20_drng {
 	struct chacha20_state chacha20;
-	struct seed_source source;
 	time_t last_seeded;
 	uint64_t generated_bytes;
 };
@@ -456,6 +503,125 @@ static int drng_chacha20_generate(struct chacha20_state *chacha20,
 	return 0;
 }
 
+static int drng_chacha20_rng_selftest(struct chacha20_drng *drng)
+{
+	int ret;
+	uint8_t outbuf[CHACHA20_KEY_SIZE * 2];
+	uint8_t seed[CHACHA20_KEY_SIZE * 2] = {
+		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
+		0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
+		0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
+		0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
+		0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
+		0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
+		0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
+		0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
+	};
+
+	/*
+	 * Expected result when ChaCha20 DRNG state is zero:
+	 *	* constants are set to "expand 32-byte k"
+	 *	* remaining state is 0
+	 * and pulling one ChaCha20 DRNG block.
+	 */
+	uint8_t expected_block[CHACHA20_KEY_SIZE] = {
+		0x76, 0xb8, 0xe0, 0xad, 0xa0, 0xf1, 0x3d, 0x90,
+		0x40, 0x5d, 0x6a, 0xe5, 0x53, 0x86, 0xbd, 0x28,
+		0xbd, 0xd2, 0x19, 0xb8, 0xa0, 0x8d, 0xed, 0x1a,
+		0xa8, 0x36, 0xef, 0xcc, 0x8b, 0x77, 0x0d, 0xc7 };
+
+	/*
+	 * Expected result when ChaCha20 DRNG state is zero:
+	 *	* constants are set to "expand 32-byte k"
+	 *	* remaining state is 0
+	 * followed by a reseed with
+	 *	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
+	 *	0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
+	 *	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
+	 *	0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
+	 *	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
+	 *	0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
+	 *	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
+	 *	0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f
+	 * and pulling two ChaCha20 DRNG blocks.
+	 */
+	uint8_t expected_twoblocks[CHACHA20_KEY_SIZE * 2] = {
+		0x80, 0xd5, 0xb1, 0x4d, 0x70, 0x5d, 0x3c, 0xa2,
+		0x23, 0x43, 0xc2, 0xe2, 0x1a, 0x4b, 0xb7, 0x29,
+		0x88, 0xed, 0x02, 0x4b, 0x4f, 0xa5, 0x52, 0xa9,
+		0xba, 0x92, 0x52, 0xcd, 0xe1, 0x0e, 0xe4, 0x87,
+		0xf9, 0xb1, 0xf6, 0xb9, 0x50, 0x3d, 0x30, 0x76,
+		0xda, 0xf8, 0x30, 0x0b, 0x0b, 0x46, 0x73, 0x6a,
+		0x9d, 0x91, 0xd3, 0xc6, 0xb1, 0xfc, 0xf3, 0x2a,
+		0xe9, 0xa3, 0x4c, 0x65, 0xd1, 0xcc, 0x37, 0x9d };
+
+	/*
+	 * Expected result when ChaCha20 DRNG state is zero:
+	 *	* constants are set to "expand 32-byte k"
+	 *	* remaining state is 0
+	 * followed by a reseed with
+	 *	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
+	 *	0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
+	 *	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
+	 *	0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
+	 *	0x20
+	 * and pulling one ChaCha20 DRNG block plus one byte.
+	 */
+	uint8_t expected_block_and_byte[CHACHA20_KEY_SIZE + 1] = {
+		0x0d, 0x7b, 0xa4, 0xec, 0x6c, 0xee, 0x5a, 0x9a,
+		0xc5, 0x6c, 0x5b, 0xa8, 0x91, 0x05, 0x71, 0xc9,
+		0x35, 0xca, 0x45, 0xdb, 0x8f, 0x10, 0xe4, 0x4a,
+		0x3b, 0x53, 0x80, 0x98, 0x82, 0x9a, 0x3b, 0x27,
+		0x5f };
+
+	/* Generate with zero state */
+	ret = drng_chacha20_generate(&drng->chacha20, outbuf,
+				     sizeof(expected_block));
+	if (ret)
+		return ret;
+	if (memcmp(outbuf, expected_block, sizeof(expected_block)))
+		return -EFAULT;
+
+	/* Clear state of DRNG */
+	memset(&drng->chacha20.key.u[0], 0, 48);
+
+	/* Reseed with 2 blocks */
+	ret = drng_chacha20_seed(&drng->chacha20, seed,
+				 sizeof(expected_twoblocks));
+	if (ret)
+		return ret;
+	ret = drng_chacha20_generate(&drng->chacha20, outbuf,
+				     sizeof(expected_twoblocks));
+	if (ret)
+		return ret;
+	if (memcmp(outbuf, expected_twoblocks, sizeof(expected_twoblocks)))
+		return -EFAULT;
+
+	/* Clear state of DRNG */
+	memset(&drng->chacha20.key.u[0], 0, 48);
+
+	/* Reseed with 1 block and one byte */
+	ret = drng_chacha20_seed(&drng->chacha20, seed,
+				 sizeof(expected_block_and_byte));
+	if (ret)
+		return ret;
+	ret = drng_chacha20_generate(&drng->chacha20, outbuf,
+				     sizeof(expected_block_and_byte));
+	if (ret)
+		return ret;
+	if (memcmp(outbuf, expected_block_and_byte,
+		   sizeof(expected_block_and_byte)))
+		return -EFAULT;
+
+	return 0;
+}
+
+static void drng_chacha20_dealloc(struct chacha20_drng *drng)
+{
+	memset_secure(drng, 0, sizeof(*drng));
+	free(drng);
+}
+
 /**
  * Allocation of the DRBG state
  */
@@ -463,7 +629,7 @@ static int drng_chacha20_alloc(struct chacha20_drng **out)
 {
 	struct chacha20_drng *drng;
 	uint32_t i, v = 0;
-	int ret;
+	int ret = 0;
 
 	if (drng_chacha20_selftest()) {
 		return -EFAULT;
@@ -477,13 +643,20 @@ static int drng_chacha20_alloc(struct chacha20_drng **out)
 
 	/* prevent paging out of the memory state to swap space */
 	ret = mlock(drng, sizeof(*drng));
-	if (ret && errno != EPERM && errno != EAGAIN)
-		return -errno;
+	if (ret && errno != EPERM && errno != EAGAIN) {
+		ret = -errno;
+		goto err;
+	}
 
 	memset(drng, 0, sizeof(*drng));
 
 	memcpy(&drng->chacha20.constants[0], "expand 32-byte k", 16);
 
+	ret = drng_chacha20_rng_selftest(drng);
+	if (ret)
+		goto err;
+
+	/* Update the state left by the self test */
 	for (i = 0; i < CHACHA20_KEY_SIZE_WORDS; i++) {
 		get_time(NULL, &v);
 		drng->chacha20.key.u[i] ^= v;
@@ -497,12 +670,10 @@ static int drng_chacha20_alloc(struct chacha20_drng **out)
 	*out = drng;
 
 	return 0;
-}
 
-static void drng_chacha20_dealloc(struct chacha20_drng *drng)
-{
-	memset_secure(drng, 0, sizeof(*drng));
-	free(drng);
+err:
+	drng_chacha20_dealloc(drng);
+	return ret;
 }
 
 /***************************** ChaCha20 DRNG API *****************************/
@@ -510,17 +681,56 @@ static void drng_chacha20_dealloc(struct chacha20_drng *drng)
 int drng_chacha20_reseed(struct chacha20_drng *drng, const uint8_t *inbuf,
 			 uint32_t inbuflen)
 {
-	uint8_t seed[CHACHA20_KEY_SIZE * OVERSAMPLINGRATE];
+	uint8_t seed[CHACHA20_KEY_SIZE * 2];
 	int ret;
+	uint32_t collected = 0;
 
-	ret = drng_get_seed(&drng->source, seed, sizeof(seed));
-	if (ret != sizeof(seed))
+	/* Entropy assumption: 1 data bit delivers one bit of entropy */
+	ret = drng_getrandom_get(seed, CHACHA20_KEY_SIZE);
+	if (ret < 0)
 		return ret;
 
-	ret = drng_chacha20_seed(&drng->chacha20, seed, sizeof(seed));
+	if (ret) {
+		collected = ret;
+
+		ret = drng_chacha20_seed(&drng->chacha20, seed,
+					 CHACHA20_KEY_SIZE);
+		if (ret)
+			return ret;
+	}
+
+	/* Entropy assumption: 2 data bits deliver one bit of entropy */
+	ret = drng_jent_get(seed, sizeof(seed));
+	if (ret < 0)
+		return ret;
+
+	if (ret) {
+		collected += ret;
+
+		ret = drng_chacha20_seed(&drng->chacha20, seed, sizeof(seed));
+		if (ret)
+			return ret;
+	}
+
+	/* Entropy assumption: 1 data bit delivers one bit of entropy */
+	ret = drng_random_get(seed, CHACHA20_KEY_SIZE);
+	if (ret < 0)
+		return ret;
+
+	if (ret) {
+		collected += ret;
+
+		ret = drng_chacha20_seed(&drng->chacha20, seed,
+					 CHACHA20_KEY_SIZE);
+		if (ret)
+			return ret;
+	}
+
 	memset_secure(seed, 0, sizeof(seed));
-	if (ret)
-		return ret;
+
+	/* Internal noise sources must have delivered sufficient information */
+	if (collected < CHACHA20_KEY_SIZE)
+		return -EFAULT;
 
 	if (inbuf && inbuflen)
 		ret = drng_chacha20_seed(&drng->chacha20, inbuf, inbuflen);
@@ -531,6 +741,13 @@ int drng_chacha20_reseed(struct chacha20_drng *drng, const uint8_t *inbuf,
 	return ret;
 }
 
+void drng_chacha20_destroy(struct chacha20_drng *drng)
+{
+	drng_jent_dealloc();
+	drng_random_dealloc();
+	drng_chacha20_dealloc(drng);
+}
+
 int drng_chacha20_init(struct chacha20_drng **drng)
 {
 	int ret = drng_chacha20_alloc(drng);
@@ -538,34 +755,19 @@ int drng_chacha20_init(struct chacha20_drng **drng)
 	if (ret)
 		return ret;
 
-	ret = drng_seedsource_alloc(&(*drng)->source);
-	if (ret)
-		goto deallocdrng;
-
 	ret = drng_chacha20_reseed(*drng, NULL, 0);
-	if (ret)
-		goto deallocsource;
+	if (ret) {
+		drng_chacha20_destroy(*drng);
+		return ret;
+	}
 
 	return 0;
-
-deallocsource:
-	drng_seedsource_dealloc(&(*drng)->source);
-deallocdrng:
-	drng_chacha20_dealloc(*drng);
-
-	return ret;
-}
-
-void drng_chacha20_destroy(struct chacha20_drng *drng)
-{
-	drng_seedsource_dealloc(&drng->source);
-	drng_chacha20_dealloc(drng);
 }
 
 int drng_chacha20_get(struct chacha20_drng *drng, uint8_t *outbuf,
 		      uint32_t outbuflen)
 {
-	time_t now;
+	time_t now = 0;
 	uint32_t nsec;
 	int ret;
 
